from collections.abc import Iterable
from typing import Union

import numpy as np
from numpy.typing import NDArray

from feature_fabrica.models import FeatureValue

NumericArray = Union[Union[NDArray[np.float32], NDArray[np.float64]], Union[NDArray[np.int32], NDArray[np.int64]]]
NumericValue = Union[np.float32, np.float64, np.int32, np.int64, float, int]

StrArray = Union[NDArray[np.str_], np.ndarray]
StrValue = Union[np.str_, str]

StrOrNumArray = Union[StrArray, NumericArray]

def broadcast_and_normalize_numeric_array(iterable: Iterable) -> NumericArray:
    # Normalize all elements to np.array
    normalized_iterable = []
    for element in iterable:  # type: ignore[union-attr]
        if not isinstance(element, np.ndarray) and not isinstance(
            element, FeatureValue
        ):
            element = np.array([element], dtype=np.float32)
        normalized_iterable.append(element)
    # Find the maximum shape among the elements
    max_shape = np.broadcast_shapes(*[elem.shape for elem in normalized_iterable])

    # Broadcast elements to the maximum shape
    broadcasted_iterable = [
        np.broadcast_to(elem, max_shape) for elem in normalized_iterable
    ]
    return broadcasted_iterable



# Define a custom exception for cyclic dependencies
class CyclicDependencyError(Exception):
    def __init__(self, loop_features):
        message = f"Cyclic dependency detected among the following features: {loop_features}. Feature dependencies must be acyclic."
        super().__init__(message)
        self.loop_features = loop_features


class FeatureNotComputedError(Exception):
    def __init__(self, f_name: str):
        message = f"The feature {f_name} must be computed before finalizing metrics."
        super().__init__(message)
        self.f_name = f_name



import numpy as np
from beartype import beartype

from feature_fabrica.transform.base import Transformation
from feature_fabrica.transform.utils import StrOrNumArray


class GroupByReduce(Transformation):
    @beartype
    def __init__(self, key_feature: str, reduce_func: Transformation, axis: int = 0):

        self.key_feature = key_feature
        self.reduce_func = reduce_func
        self.axis = axis
    @beartype
    def execute(self, data: StrOrNumArray) -> StrOrNumArray:
        """Groups data array by key_feature values and applies a reduction transformation.

        Assigns reduced values to the original positions to preserve the original shape.
        """
        assert self.reduce_func.expects_data
        #assert isinstance(self.key_feature, FeatureValue), "key_feature must be an existing feature!"
        assert self.key_feature.shape == data.shape # type: ignore[attr-defined]
        key_feature = self.key_feature

        sorted_idxs = key_feature.argsort() # type: ignore[attr-defined]
        data_sorted = data[sorted_idxs]
        key_feature_sorted = key_feature[sorted_idxs]

        data_aggregated =  np.split(data_sorted, np.unique(key_feature_sorted, return_index=True)[1][1:])

        data_aggregated = np.array(data_aggregated, dtype=data.dtype)
        data_reduced = self.reduce_func.execute(data_aggregated)

        # Use np.unique to find the unique key values and the counts for each
        _, inverse_indices = np.unique(key_feature, return_inverse=True)

        # Use np.repeat to assign the reduced values to the original positions
        result = data_reduced[inverse_indices]

        return result
